##############################################################################
#
# Copyright (c) 2009 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################

from __future__ import absolute_import

import logging

from zope.interface import implementer

from .interfaces import IScriptRunner

log = logging.getLogger(__name__)


@implementer(IScriptRunner)
class ScriptRunner(object):

    # script_vars contains replacements for parts of scripts.
    # These are correct for PostgreSQL and MySQL but not for Oracle.
    # The primary purpose of this originally was to compensate for different
    # parameter passing styles in the DB-API. It has also been used for some
    # SQL syntax/datatype differences.
    #
    # TODO: Make a clean separation between syntax, and DB-API substitutions.
    # Use `script_vars` (or a better name) for parameters, and
    # `format_vars` (or a better name) for syntax? Or just wrap this up in a
    # query_property(..., formatted=True)?
    script_vars = {
        'TRUE':         'TRUE',
        'FALSE':        'FALSE',
        'TRUNCATE':     'TRUNCATE TABLE',
        'oid':          '%(oid)s',
        'tid':          '%(tid)s',
        'pack_tid':     '%(pack_tid)s',
        'undo_tid':     '%(undo_tid)s',
        'self_tid':     '%(self_tid)s',
        'min_tid':      '%(min_tid)s',
        'max_tid':      '%(max_tid)s',
        'INNER_ORDER_BY': 'ORDER BY zoid',
    }

    format_vars = {
    }

    def new_instance(self):
        return type(self)()

    def with_format_vars(self, **new_vars):
        inst = self.new_instance()
        inst.format_vars = dict(self.format_vars)
        inst.format_vars.update(new_vars)
        return inst

    def run_script_stmt(self, cursor, generic_stmt, generic_params=()):
        """Execute a statement from a script with the given parameters.

        params should be either an empty tuple (no parameters) or
        a map.

        The input statement is generic and needs to be transformed
        into a database-specific statement.
        """
        __traceback_info__ = generic_stmt, self.format_vars
        stmt = generic_stmt.format(**self.format_vars)
        if '%(' in stmt and ')s' in stmt:
            __traceback_info__ = stmt, self.script_vars
            stmt = stmt % self.script_vars
        __traceback_info__ = stmt
        try:
            cursor.execute(stmt, generic_params)
        except:
            log.warning("script statement failed: %r; parameters: %r",
                        stmt, generic_params)
            raise

    def run_script(self, cursor, script, params=()):
        """Execute a series of statements in the database.

        params should be either an empty tuple (no parameters) or
        a map.

        The statements are transformed by run_script_stmt
        before execution.
        """
        lines = []
        for line in script.split('\n'):
            line = line.strip()
            if not line or line.startswith('--'):
                continue
            if line.endswith(';'):
                line = line[:-1]
                lines.append(line)
                stmt = '\n'.join(lines)
                self.run_script_stmt(cursor, stmt, params)
                lines = []
            else:
                lines.append(line)
        if lines:
            stmt = '\n'.join(lines)
            self.run_script_stmt(cursor, stmt, params)

    def run_many(self, cursor, stmt, items):
        """Execute a statement repeatedly.  Items should be a list of tuples.

        stmt should use '%s' parameter format.
        """
        cursor.executemany(stmt, items)



# -*- coding: utf-8 -*-
"""
Helper implementations for MVCC.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import threading
from logging import DEBUG as LDEBUG

from zope import interface

from BTrees import family64

from .interfaces import IMVCCDatabaseCoordinator
from .interfaces import IDetachableMVCCDatabaseViewer

logger = __import__('logging').getLogger(__name__)

@interface.implementer(IDetachableMVCCDatabaseViewer)
class DetachableMVCCDatabaseViewer(object):
    __slots__ = (
        'highest_visible_tid',
        'detached',
    )

    def __init__(self):
        self.highest_visible_tid = None
        self.detached = False


@interface.implementer(IMVCCDatabaseCoordinator)
class DetachableMVCCDatabaseCoordinator(object):
    """
    Simple implementation of :class:`IMVCCDatabaseCoordinator`
    that works with :class:`relstorage.interfaces.IDetachableMVCCDatabaseViewer`
    objects.

    We keep hard references to our viewers, so if they reference us there
    could be a cycle. Viewers must be hashable.

    The ``highest_visible_tid`` and ``detached`` values of the viewer
    must *only* be managed through this object.
    """

    maximum_highest_visible_tid = None
    minimum_highest_visible_tid = None

    def __init__(self):
        # Manipulations of metadata must be locked.
        # We don't always hold the lock; we rely on primitive operations of
        # the set() in _registered_viewers to be atomic.
        self._lock = threading.RLock()
        # {tid: {viewer, ...}} of objects not detached and not None
        self._by_tid = family64.UO.Bucket()
        self._registered_viewers = set()
        self.is_registered = self._registered_viewers.__contains__
        self.log = logger.log

    @property
    def _viewer_count_at_min(self):
        # Testing.
        if not self._registered_viewers or not self.minimum_highest_visible_tid:
            return 0
        viewers = self._by_tid.values()[0]
        return len(viewers)

    def register(self, viewer):
        with self._lock:
            if self.is_registered(viewer):
                return
            self._registered_viewers.add(viewer)
            if viewer.detached:
                return

            hvt = viewer.highest_visible_tid
            if hvt is None:
                return

            __traceback_info__ = hvt, viewer
            self._by_tid.setdefault(hvt, set()).add(viewer)
            self.minimum_highest_visible_tid = self._by_tid.minKey()
            self.maximum_highest_visible_tid = self._by_tid.maxKey()

    def unregister(self, viewer):
        with self._lock:
            if not self.is_registered(viewer):
                return

            self._registered_viewers.remove(viewer)
            if not self._registered_viewers:
                self.minimum_highest_visible_tid = None
                self.maximum_highest_visible_tid = None
                self._by_tid.clear()
                return

            self.__viewer_does_not_matter(viewer)

    def __set_tids(self):
        by_tid = self._by_tid
        if by_tid:
            self.minimum_highest_visible_tid = by_tid.minKey()
            self.maximum_highest_visible_tid = by_tid.maxKey()
        else:
            self.minimum_highest_visible_tid = None
            self.maximum_highest_visible_tid = None


    def __viewer_does_not_matter(self, viewer):
        # Because it was unregistered or because it
        # was detached.
        hvt = viewer.highest_visible_tid
        by_tid = self._by_tid
        if by_tid and hvt:
            viewers = by_tid.get(hvt)
            if viewers:
                viewers.discard(viewer)
                if not viewers:
                    del by_tid[hvt]
                self.__set_tids()

    def clear(self):
        with self._lock:
            self._registered_viewers.clear()
            self._by_tid.clear()
            self.__set_tids()

    def detach(self, viewer):
        """
        Cause the viewer to become detached.
        """
        with self._lock:
            if not self.is_registered(viewer):
                return

            viewer.detached = True
            self.__viewer_does_not_matter(viewer)

    def detach_all(self):
        with self._lock:
            for viewer in self._registered_viewers:
                viewer.detached = True
            self._by_tid.clear()
            self.__set_tids()

    def change(self, viewer, new_hvt):
        """
        Cause the viewer to have a new ``highest_visible_tid``,
        which can be greater, less, or equal to the current HVT,
        or None.

        If the *viewer* was previously detached, it is now attached.
        """
        with self._lock:
            if not self.is_registered(viewer):
                return

            viewer.detached = False

            old_hvt = viewer.highest_visible_tid
            viewer.highest_visible_tid = new_hvt
            by_tid = self._by_tid
            if old_hvt:
                viewers = by_tid.get(old_hvt)
                if viewers:
                    viewers.discard(viewer)
                    if not viewers:
                        del by_tid[old_hvt]
            if new_hvt:
                by_tid.setdefault(new_hvt, set()).add(viewer)

            self.__set_tids()

    def viewers_at_or_before(self, tid):
        """
        Return all the viewers with tids at least as old as the
        given tid.

        Passing the value from ``minumum_highest_visible_tid`` is always safe,
        even if that value is None. If that value is None, it's because we
        have no viewers, or the viewers we do have haven't looked at the
        database; they'll be ignored.

        Viewers that are already explicitly detached are also ignored.
        """
        with self._lock:
            by_tid = self._by_tid
            if not by_tid:
                return ()
            sets_before = by_tid.values(max=tid, excludemax=False)
            return set().union(*sets_before) if sets_before else ()

    def viewers_at_minimum(self):
        """
        Return all the viewers viewing the ``minimum_highest_visible_tid``.

        If that is None, this is the empty set.
        """
        with self._lock:
            if self._by_tid:
                return self._by_tid.values()[0]
            return ()

    def detach_viewers_at_minimum(self):
        """
        Cause all the viewers at the minimum, if any, to be detached.
        """
        with self._lock:
            if self._by_tid:
                min_tid = self._by_tid.minKey()
                at_min = self._by_tid.pop(min_tid)
                if at_min:
                    self.log(
                        LDEBUG,
                        "Detaching %d viewers at oldest TID 0x%x",
                        len(at_min), min_tid,
                    )
                    for viewer in at_min:
                        viewer.detached = True
                self.__set_tids()



##############################################################################
#
# Copyright (c) 2009 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import logging
import struct
import threading
import time

from ZODB.utils import p64
from ZODB.utils import z64

log = logging.getLogger(__name__)


class ZEOTracer(object):
    # Knows how to write ZEO trace files.

    def __init__(self, trace_file):
        self._trace_file = trace_file
        self._lock = threading.Lock()

        # closure variables are faster than self/global dict lookups
        # (going off example in ZEO code; in one test locally this gets us a
        # ~15% improvement)
        _now = time.time
        _pack = struct.Struct(">iiH8s8s").pack
        _trace_file_write = trace_file.write
        _p64 = p64
        _z64 = z64
        _int = int
        _len = len

        def trace(code, oid_int=0, tid_int=0, end_tid_int=0, dlen=0, now=None):
            # This method was originally part of ZEO.cache.ClientCache. The below
            # comment is verbatim:
            # The code argument is two hex digits; bits 0 and 7 must be zero.
            # The first hex digit shows the operation, the second the outcome.
            # ...
            # Note: when tracing is disabled, this method is hidden by a dummy.
            encoded = (dlen << 8) + code
            tid = _p64(tid_int) if tid_int else _z64
            end_tid = _p64(end_tid_int) if end_tid_int else _z64
            oid = b'' if not oid_int else _p64(oid_int)

            now = now or _now()
            try:
                _trace_file_write(
                    _pack(
                        _int(now), encoded, _len(oid), tid, end_tid) + oid,
                )
            except: # pragma: no cover
                log.exception("Problem writing trace info for %r at tid %r and end tid %r",
                              oid, tid, end_tid)
                raise

        self._trace = trace

    def trace(self, code, oid_int=0, tid_int=0, end_tid_int=0, dlen=0):
        with self._lock:
            self._trace(code, oid_int, tid_int, end_tid_int, dlen)

    def trace_store_current(self, tid_int, state_oid_iter):
        # As a locking optimization, we accept this in bulk
        # Theoretically this could be any iterable, but
        # we only work with the one defined in storage_cache.
        with self._lock:
            now = time.time()
            for startpos, endpos, oid_int, _prev_tid_int in state_oid_iter.items():
                self._trace(0x52, oid_int, tid_int, dlen=endpos - startpos, now=now)

    def close(self):
        self._trace_file.close()
        del self._trace



# -*- coding: utf-8 -*-
##############################################################################
#
# Copyright (c) 2009 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""
IOIDAllocator implementations.
"""

from __future__ import absolute_import

from zope.interface import implementer


from ..interfaces import IOIDAllocator
from ..oidallocator import AbstractTableOIDAllocator
from .locker import lock_timeout



@implementer(IOIDAllocator)
class MySQLOIDAllocator(AbstractTableOIDAllocator):

    # https://dev.mysql.com/doc/refman/5.7/en/example-auto-increment.html
    # "Updating an existing AUTO_INCREMENT column value in an InnoDB
    # table does not reset the AUTO_INCREMENT sequence as it does for
    # MyISAM and NDB tables."
    #
    # Thus we must do a pure INSERT, not a REPLACE or UPDATE..
    #
    # Alternatively, ALTER TABLE ... AUTO_INCREMENT = X;
    # But: That behaves different by version (persistent on 8, non-persistent
    # on 5.7; that's probably ok.) However, it causes an auto-commit, which
    # is certainly not desired as this is used by a store connection.
    #
    # https://dev.mysql.com/doc/refman/5.7/en/innodb-auto-increment-handling.html#innodb-auto-increment-lock-modes
    # In InnoDB, it must be indexed as a leading column.
    #
    # "In all lock modes (0, 1, and 2), if a transaction that
    # generated auto-increment values rolls back, those auto-increment
    # values are 'lost'."
    #
    # "If you delete the row containing the maximum value for an
    # AUTO_INCREMENT column, the value is not reused for a MyISAM or
    # InnoDB table. If you delete all rows in the table with DELETE
    # FROM tbl_name (without a WHERE clause) in autocommit mode, the
    # sequence starts over for all storage engines except InnoDB and
    # MyISAM."
    #
    # Regarding `TRUNCATE TABLE`:
    #
    # "Any AUTO_INCREMENT value is reset to its start value. This is
    # true even for MyISAM and InnoDB, which normally do not reuse
    # sequence values."
    def _set_min_oid_from_range(self, cursor, n):
        # A simple statement like the following can easily deadlock:
        #
        # INSERT INTO new_oid (zoid)
        # SELECT %s
        # WHERE %s > (SELECT COALESCE(MAX(zoid), 0) FROM new_oid)
        #
        #
        # Session A: new_oid() -> Lock 1
        # Session B: new_oid() -> Lock 2 (row 1 is invisible)
        # Session A: new_oid() -> Lock 3 (row 2 is invisible)
        # Session B: set_min_oid(2) -> Hang waiting for lock
        # Session A: new_oid() -> Lock 4: Deadlock, Session B rolls back.
        #
        # Partly this is because MAX() is local to the current session.
        # We deal with this by using a stored procedure to efficiently make
        # multiple queries.
        self.driver.callproc_no_result(cursor, 'set_min_oid(%s)', (n,))

   # Notes on new_oids:

    # Generate a new auto_incremented ID. This will never conflict
    # with any other session because generated IDs are guaranteed
    # to be unique. However, the DELETE statement may interfere
    # with other sessions and lead to deadlock; this is true even
    # when using the 'DELETE IGNORE'. The INSERT statement takes out an
    # exclusive lock on the PRIMARY KEY index.
    #
    # Session A: INSERTS 2000. -> lock 2000
    # Session B: INSERTS 3000. -> lock 3000
    # Session B: DELETE < 3000 -> wants lock on 2000; locks everything under.
    # Session A: DELETE < 2000 -> hold lock on 2000; wants to lock things under.
    #
    # ORDER BY zoid (ASC or DESC) just runs into more lock issues reported for
    # individual rows.
    #
    # Using the single-row update example provided by
    # https://dev.mysql.com/doc/refman/5.7/en/innodb-locking-reads.html
    # also runs into deadlocks (naturally).
    #
    # Our solution is to just let rows build up if the delete fails. Eventually
    # a GC, which happens at startup, will occur and hopefully get most of them.

    # cursor.lastrowid is used in the superclass; that's a DB-API
    # extension. Fortunately, all supported drivers implement it. (In
    # the past we used cursor.connection.insert_id(), which was
    # specific to MySQLdb and PyMySQL.) 'SELECT LAST_INSERT_ID()' is
    # the pure-SQL way to do this.

    def _timeout(self, cursor, batch_timeout):
        return lock_timeout(cursor, batch_timeout)



##############################################################################
#
# Copyright (c) 2009 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function


from hamcrest import assert_that
from nti.testing.matchers import validly_provides

# The overhead of cache values, and thus how much fits in a ring or the
# cache, depends on 32 or 64 bit, whether or not we copy
# bytes into native code std::string, and how the compiler lays out
# the objects. We only copy strings on PyPy, but we have little control
# over the object layout, especially with the various MSVC compilers
# we have to deal with. So that explains the tests that have a range of sizes.

from relstorage.tests import TestCase
from relstorage.cache import interfaces
from . import Cache

class GenerationTests(TestCase):

    def _makeCache(self, limit):
        from . import Cache as BaseCache
        return BaseCache(limit)

    def _makeOne(self, limit):
        return self._makeCache(limit).eden

    def test_bool(self):
        cache = self._makeCache(100)
        lru = cache.eden
        self.assertFalse(lru)

        cache[1] = (b'', 0)
        self.assertTrue(lru)
        self.assertEqual(1, len(lru))

        del cache[1]
        self.assertFalse(lru)

class EdenTests(TestCase):

    def _makeOne(self, limit):
        from . import Cache as BaseCache
        return BaseCache(limit)

    def test_add_MRUs_empty(self):
        lru = self._makeOne(100)
        self.assertEqual((), lru.add_MRUs([]))

    def test_add_MRUs_too_many(self):
        lru = self._makeOne(1000)
        too_many = [(i, (b'a' * i, 0, 0, 1)) for i in range(50)]
        # They just exceed the limit
        added = lru.add_MRUs(too_many)
        # Much less got added
        self.assertGreaterEqual(len(added), 7)
        self.assertLessEqual(len(added), 9)

class NoOverheadSizeCache(Cache):
    def __init__(self, byte_limit):
        Cache.__init__(self, byte_limit)
        self.base_size = self.weight
        self[0] = (b'', 0)
        self.entry_size = self[0].weight
        del self[0]

    @property
    def weight(self):
        weight = super(NoOverheadSizeCache, self).size
        weight -= self.base_size
        weight -= self.entry_size * len(self)
        return weight

    @property
    def size(self):
        return self.weight

    def __getitem__(self, oid):
        entry = Cache.__getitem__(self, oid)
        if entry is not None:
            return self.get_item_with_tid(oid, entry.tid)


class GenericLRUCacheTests(TestCase):
    """
    Generic LRU caching tests that can be applied to any
    LRU implementation, using the kind of keys and
    values that we actually use: ``(oid_int, tid_int)`` and
    ``(state_bytes, tid_int)``.
    """

    def _getClass(self):
        return NoOverheadSizeCache

    def _makeOne(self, limit, kind=None):
        kind = kind or self._getClass()
        return kind(limit)

    def _getIface(self):
        return interfaces.ILRUCache

    def test_implements(self):
        cache = self._makeOne(100)
        assert_that(cache,
                    validly_provides(self._getIface()))
        return cache

    def test_eden_implements(self):
        cache = self._makeOne(100)
        assert_that(cache.eden,
                    validly_provides(interfaces.IGeneration))

    def test_item_implements(self):
        cache = self._makeOne(20)
        cache[1] = (b'', 0)
        entrya = cache[1]
        assert_that(entrya, validly_provides(interfaces.ILRUEntry))

    def test_add_too_many(self):
        class _Cache(self._getClass()):
            pass
        cache = _Cache(20 + _Cache(20).base_size + (_Cache(20).entry_size * 2))

        entry_count = 10
        entries = cache.add_MRUs(list(reversed([
            # oid, state, frozen, frequency
            (x, (b'abcde', 0, False, x))
            for x
            in range(1, entry_count)
        ])))

        self.assertEqual(
            [5 + cache.entry_size] * len(entries),
            [e.weight for e in entries])
        self.assertLessEqual(
            cache.weight,
            cache.limit
        )
        self.assertEqual(
            [e.key for e in entries],
            [e.frequency for e in entries])
        self.assertEqual(
            [8, 7, 6, 5],
            [e.key for e in entries]
        )
        self.assertEqual(4, len(cache))
        return cache

    def test_age(self):
        cache = self._makeOne(100)
        base_size = cache.base_size
        entry_size = cache.entry_size
        cache = self._getClass()(100 + base_size + entry_size)


        entries = cache.add_MRUs([
            (1, (b'abcde', 0, False, 1)),
            (2, (b'abcde', 0, False, 1)),
            (3, (b'abcde', 0, False, 1)),
            (0, (b'abcde', 0, False, 1)),
        ])

        self.assertIn(
            [e.key for e in entries],
            ([1, 2, 3, 0], [2, 3, 0])
        )

        for _ in range(4):
            for e in entries:
                _ = cache[e.key]

        freqs = [e.frequency for e in cache.values()]
        self.assertEqual([5] * len(entries), freqs)

        # By half each time
        cache.age_frequencies()
        freqs = [e.frequency for e in cache.values()]
        self.assertEqual([2] * len(entries), freqs)
        return cache


    def test_delete(self):
        cache = self._makeOne(20)
        cache[1] = (b'abc', 0)
        self.assertIn(1, cache)
        self.assertEqual(1, len(cache))
        self.assertEqual(3, cache.size)
        self.assertEqual(cache[1], (b'abc', 0))
        self.assertEqual(list(cache), [(1, 0)])
        del cache[1]
        self.assertNotIn(1, cache)
        self.assertEqual(0, len(cache))
        self.assertEqual(0, cache.size)
        self.assertIsNone(cache[1])
        self.assertEqual(list(cache), [])

    def test_entries(self):
        cache = self._makeOne(20)
        cache[1] = (b'abc', 0)
        entries = list(cache.values())
        self.assertEqual(1, len(entries))
        entry = entries[0]
        assert_that(entry, validly_provides(interfaces.ILRUEntry))
        self.assertEqual(1, entry.key)
        self.assertEqual(b'abc', entry.value)
        self.assertEqual(1, entry.frequency)

        # Getting it again updates its frequency, not
        # necessarily on the same object though.
        self.assertIsNotNone(cache[1])
        entries = list(cache.values())
        self.assertEqual(1, len(entries))
        entry = entries[0]
        self.assertEqual(1, entry.key)
        self.assertEqual(b'abc', entry.value)
        self.assertEqual(2, entry.frequency)

    def test_add_too_many_MRUs_works_aronud_big_entry(self):
        cache = self._getClass()(20)
        base_size = cache.base_size
        entry_size = cache.entry_size
        cache = self._getClass()(40 + base_size + entry_size)

        entries = cache.add_MRUs([
            (1, (b'abc', 0, False, 1)),
            # This entry itself will fit nowhere
            (2, (b'12345678901234567890' * 20, 0, False, 1)),
            (3, (b'bcd', 0, False, 1)),
            (4, (b'cde', 0, False, 1)),
            (5, (b'dehi', 0, False, 1)),
            (6, (b'edghijkl', 0, False, 1)),
        ])
        self.assertGreaterEqual(len(cache), 3)
        self.assertLessEqual(len(cache), 4)

        self.assertIn(
            [e.key for e in entries],
            ([1, 3, 4, 5], [3, 4, 5]))
        return cache


class GenericGenerationalLRUCacheTests(GenericLRUCacheTests):
    """
    Tests for any generational LRU cache.
    """

    def test_implements(self):
        cache = super(GenericGenerationalLRUCacheTests, self).test_implements()
        assert_that(cache.eden,
                    validly_provides(interfaces.IGeneration))
        assert_that(cache.protected,
                    validly_provides(interfaces.IGeneration))
        assert_that(cache.probation,
                    validly_provides(interfaces.IGeneration))


    def test_bad_generation_index_attribute_error(self):
        cache = self._makeOne(20)
        # Check proper init
        getattr(cache.generations[1], 'limit')
        getattr(cache.generations[2], 'limit')
        getattr(cache.generations[3], 'limit')

        # Gen 0 should be missing
        with self.assertRaisesRegex(AttributeError,
                                    "Generation 0 has no attribute 'on_hit'"):
            cache.generations[0].on_hit()

    def test_add_MRUs_reject_sets_sentinel_values(self):
        # When we find an item that completely fills the cache,
        # all the rest of the items are marked as rejected.
        cache = self._getClass()(20)
        base_size = cache.base_size
        entry_size = cache.entry_size
        cache = self._getClass()(20 + base_size + entry_size)

        added_entries = cache.add_MRUs([
            # over fill eden with item of size 15
            (1, (b'012345678901234' * 20, 0, False, 1)),
            # 1 goes to protected, filling it. eden is also over full with 2. probation is empty
            (2, (b'012', 0, False, 1)),
            # 3 fills eden, bumping 2 to probation. But probation is actually overfull now
            # so we'd like to spill something if we could (but we can't.)
            (3, (b'0', 0, False, 1)),
            # 4 should never be added because it won't fit anywhere.
            (4, (b'ee', 0, False, 1)),
        ])

        def keys(x):
            return [e.key for e in x]

        self.assertEqual(keys(cache.protected), [3, 2])
        self.assertEqual(keys(cache.probation), [])

        self.assertEqual(keys(cache.eden), [4])
        self.assertEqual(
            [2, 3, 4],
            [e.key for e in added_entries])

        self.assertEqual(3, len(added_entries))
        self.assertEqual(3, len(cache))
        self.assertEqual(3, len(list(cache)))

class CFFICacheTests(TestCase):
    """
    Tests that are specific to the CFFI implementation
    of the cache.

    These can use arbitrary keys and values.
    """

    def _getClass(self):
        return NoOverheadSizeCache

    def _makeOne(self, limit, kind=None):
        self.skipTest("Weights not supported")
        kind = kind or self._getClass()
        return kind(limit,
                    key_weight=self.key_weight,
                    value_weight=self.value_weight)

    def key_weight(self, k):
        return len(k)

    def value_weight(self, v):
        return len(v)

    def test_free_reuse(self):
        cache = self._makeOne(20)
        lru = cache.protected
        self.assertEqual(lru.limit, 16)
        entrya = lru.add_MRU('a', b'')
        entryb = lru.add_MRU('b', b'')
        entryc = lru.add_MRU('c', b'1')
        entryd = lru.add_MRU('d', b'1')
        evicted = lru.update_MRU(entryb, b'1234567890')
        self.assertEqual(evicted, ())
        # Not changing the size is just a hit, it doesnt't
        # evict anything.
        evicted = lru.update_MRU(entryb, b'1234567890')
        self.assertEqual(evicted, ())
        evicted = lru.update_MRU(entryc, b'1234567890')

        # a and d were evicted and placed on the freelist
        self.assertEqual(entrya.key, None)
        self.assertEqual(entrya.value, None)
        self.assertEqual(entryd.key, None)
        self.assertEqual(entryd.key, None)

        self.assertEqual(evicted,
                         [('a', b''),
                          ('d', b'1')])
        self.assertEqual(2, len(lru.node_free_list))

        lru.add_MRU('c', b'1')
        self.assertEqual(1, len(lru.node_free_list))



# -*- coding: utf-8 -*-
"""
Utility functions and base classes.

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from copy import copy as stdlib_copy

from .interfaces import IBindParam

def copy(obj):
    """
    Make a shallow copy of the object, ignoring any volatile attributes.
    """
    new = stdlib_copy(obj)
    volatile = [k for k in vars(new) if k.startswith('_v')]
    for k in volatile:
        delattr(new, k)
    return new

class Resolvable(object):

    __slots__ = ()

    def resolve_against(self, table):
        # pylint:disable=unused-argument
        return self



class Columns(object):
    """
    Grab bag of columns.
    """

    def __init__(self, columns):
        cs = []
        for c in columns:
            try:
                col_name = c.alias or c.name
            except AttributeError:
                pass
            else:
                setattr(self, col_name, c)
            cs.append(c)
        self._columns = tuple(cs)

    def __bool__(self):
        return bool(self._columns)

    __nonzero__ = __bool__

    def __getattr__(self, name):
        # Here only so that pylint knows this class has a set of
        # dynamic attributes.
        raise AttributeError("Column list %s does not include %s" % (
            self._col_list() or "()",
            name
        ))

    def has_column(self, name):
        return hasattr(self, name)

    def __contains__(self, column):
        return column in self._columns

    def is_subset(self, other):
        return all(c in other for c in self)

    def __getitem__(self, ix):
        return self._columns[ix]

    def __iter__(self):
        return iter(self._columns)

    def _col_list(self):
        return ','.join(str(c) for c in self._columns)

    def __compile_visit__(self, compiler):
        compiler.visit_csv(self._columns)

    def has_bind_param(self):
        return any(
            IBindParam.providedBy(c)
            for c in self._columns
        )

    def is_ambiguous(self, column):
        """
        Does the *column* name appear more than once?

        If so it needs to be qualified.
        """
        count = 0
        for c in self._columns:
            if c.name == column.name:
                count += 1
        return count > 1

    def as_select_list(self):
        from .select import _SelectColumns
        return _SelectColumns(self._columns)

    def as_names(self):
        return {c.name for c in self}

    def as_dict(self):
        # This fails if there are duplicates.
        return {col.name: col for col in self._columns}

    def union(self, other):
        """
        Discards duplicates.
        """
        od = other.as_dict()
        od.update(self.as_dict())
        return type(self)(
            sorted(od.values(), key=lambda c: c.name))

    def dup_union(self, other, combining=()):
        """
        Preserves duplicates, with the exception of columns in *combining*.
        """
        seen_combining = set()
        columns = []

        for c in self._columns + other._columns:
            if c.name in combining:
                if c.name not in seen_combining:
                    columns.append(c)
                    seen_combining.add(c.name)
            else:
                columns.append(c)
        return type(self)(columns)

    def intersection(self, other):
        col_names = set(other.as_dict()).intersection(set(self.as_dict()))
        col_names = sorted(col_names)
        return type(self)(
            getattr(self, c)
            for c in col_names
        )

    def __repr__(self):
        return '<%s %s>' % (
            type(self).__name__,
            ', '.join(repr(c) for c in self)
        )

    def __eq__(self, other):
        try:
            return self.as_dict() == other.as_dict()
        except AttributeError:
            return NotImplemented



##############################################################################
#
# Copyright (c) 2008 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Tests of relstorage.adapters.oracle"""
import logging
import os
import sys
import unittest

from relstorage.adapters.oracle import OracleAdapter

from .util import AbstractTestSuiteBuilder

class OracleAdapterMixin(object):

    def make_adapter(self, options, db=None):
        dsn = os.environ.get('ORACLE_TEST_DSN', 'XE')
        if db is None:
            if self.keep_history:
                db = self.base_dbname
            else:
                db = self.base_dbname + '_hf'
        return OracleAdapter(
            user=db,
            password='relstoragetest',
            dsn=dsn,
            options=options,
        )

    def get_adapter_class(self):
        return OracleAdapter

    def __get_adapter_zconfig_dsn(self):
        dsn = os.environ.get('ORACLE_TEST_DSN', 'XE')
        return dsn

    def get_adapter_zconfig(self):
        if self.keep_history:
            dbname = self.base_dbname
        else:
            dbname = self.base_dbname + '_hf'
        return u"""
        <oracle>
            driver %s
            user %s
            password relstoragetest
            dsn %s
        </oracle>
        """ % (
            self.driver_name,
            dbname,
            self.__get_adapter_zconfig_dsn()
        )

    def verify_adapter_from_zconfig(self, adapter):
        if self.keep_history:
            dbname = self.base_dbname
        else:
            dbname = self.base_dbname + '_hf'

        self.assertEqual(adapter._user, dbname)
        self.assertEqual(adapter._password, 'relstoragetest')
        self.assertEqual(adapter._dsn, self.__get_adapter_zconfig_dsn())
        self.assertEqual(adapter._twophase, False)

    def get_adapter_zconfig_replica_conf(self):
        # import tempfile
        # dsn = self.__get_adapter_zconfig_dsn()
        # fd, replica_conf = tempfile.mkstemp('.conf', 'rstest_oracle_replica')
        # self.addCleanup(os.remove, replica_conf)
        # os.write(fd, dsn.encode("ascii"))
        # os.close(fd)
        # return replica_conf
        return None


class OracleTestSuiteBuilder(AbstractTestSuiteBuilder):

    __name__ = 'Oracle'

    def __init__(self):
        from relstorage.adapters.oracle import drivers
        super(OracleTestSuiteBuilder, self).__init__(
            drivers,
            OracleAdapterMixin,
        )

    def _compute_large_blob_size(self, use_small_blobs):
        if use_small_blobs:
            large_blob_size = 1024 * 1024 * 10
        else:
            large_blob_size = min(sys.maxsize, 1<<32)
        return large_blob_size


def test_suite():
    return OracleTestSuiteBuilder().test_suite()

if __name__ == '__main__':
    logging.basicConfig()
    logging.getLogger("zc.lockfile").setLevel(logging.CRITICAL)
    unittest.main(defaultTest="test_suite")



# -*- coding: utf-8 -*-
##############################################################################
#
# Copyright (c) 2009 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

# Perhaps pylint gets confused because of the mysqldb module in this package?
# pylint:disable=no-name-in-module,import-error
from MySQLdb.connections import Connection as BaseConnection
from MySQLdb.cursors import SSCursor

from ...drivers import GeventConnectionMixin
from . import IterateFetchmanyMixin

class Cursor(IterateFetchmanyMixin,
             SSCursor):
    """
    ``SSCursor.execute`` and ``SSCursor.callproc`` call
    ``SSCursor._query``; ``self_query`` and ``self.nextset`` both call
    ``self._get_result``. That produces a ``ResultObject``, the
    constructor of which calls the libmysql function
    ``mysql_use_result()`` and ``mysql_more_results``. The ``ResultObject`` is stored in
    ``self._result`` and has a ``has_next`` property that's statically
    determined (without reading from the network) saying whether
    there is another result set to read by ``mysql_more_results``.

    Before ``SSCursor.nextset()`` calls ``_get_result`` it calls
    the Connection's ``next_result`` method to advance the connection
    state. That method may block.
    """

    # Somewhat surprisingly, if we just wait on read,
    # we end up blocking forever. This is because of the buffers
    # maintained inside the MySQL library: we might already have the
    # rows that we need buffered.

    # Blocking on write is pointless: by definition
    # we're here to read results so we can always
    # write. That just forces us to take a trip around
    # the event loop for no good reason.

    # Therefore, our best option to periodically yield
    # is to explicitly invoke gevent.sleep(). Without
    # any time given, it will yield to other ready
    # greenlets; only sometimes will it force a trip
    # around the event loop.

    # TODO: But if we're actually going to need to read data,
    # we want to wait until it arrives. Can we get that info
    # somehow? Can we develop a heuristic?

    # Somewhat amusingly, the low-levels of libmysqlclient use
    # kevent to do waiting: mysql_real_query -> cli_read_query_result ->
    # my_net_read -> net_read_packet -> via_read_buff -> vio_read
    # -> vio_io_wait -> kevent
    #
    # We do this *after* fetching results, not before, just in case
    # we only needed to make one fetch.

    def execute(self, query, args=None):
        # The super implementation likes to call nextset() in a loop to
        # scroll through anything left dangling. But we promise not to do that
        # so we can save some effort.
        self.nextset = _noop # pylint:disable=attribute-defined-outside-init
        try:
            return SSCursor.execute(self, query, args)
        finally:
            del self.nextset

    def fetchall(self):
        if self.connection.is_in_critical_phase():
            return SSCursor.fetchall(self)
        sleep = self.connection.gevent_sleep
        result = []
        fetch = self.fetchmany # calls _fetch_row with the arraysize
        while 1:
            # Even if self.rowcount is 0 we must still call
            # or we get the connection out of sync.
            rows = fetch()
            if not rows:
                break
            result.extend(rows)
            if self.rownumber == self.rowcount:
                # Avoid a useless extra trip at the end.
                break
            sleep()

        return result

def _noop(_duration=None):
    "Does nothing"

class AbstractConnection(GeventConnectionMixin):
    default_cursor = Cursor

    # pylint:disable=method-hidden

    def query(self, query):
        # From the mysqlclient implementation:
        # "Since _mysql releases the GIL while querying, we need immutable buffer"
        # RelStorage never passes a bytearray, but third-parties have
        # been known to use this connection implementation.
        if isinstance(query, bytearray):
            query = bytes(query)

        # Prior to mysqlclient 1.4, there was a 'waiter' Connection
        # argument that could be used to do this, but it was removed.
        # So we implement it ourself.
        self.gevent_wait_write()
        self.send_query(query) # pylint:disable=no-member

        self.gevent_wait_read()
        self.read_query_result() # pylint:disable=no-member

    # The default implementations of 'rollback' and 'commit' use only
    # C API functions `mysql_rollback` and `mysql_commit`; it doesn't
    # touch any internal state. Those in turn simply call
    # `mysql_real_query("rollback")` and `mysql_real_query("commit")`.
    # That's a synchronous function that waits for the result to be
    # ready. We don't want to block like that (commit could
    # potentially take some time.)

    def rollback(self):
        self.query(b'ROLLBACK')

    def commit(self):
        self.query(b'COMMIT')

    ###
    # Critical sections
    ###
    def enter_critical_phase_until_transaction_end(self):
        # Must be idempotent
        self.commit = self._critical_commit
        self.rollback = self._critical_rollback
        self.query = self._critical_query
        self.gevent_sleep = _noop

    def is_in_critical_phase(self):
        return 'gevent_sleep' in self.__dict__

    def __exit_critical_phase(self):
        # Unconditional, assumes we're in one.
        del self.rollback
        del self.commit
        del self.query
        del self.gevent_sleep

    def exit_critical_phase(self):
        # Public, may or may not be in one.
        if self.is_in_critical_phase():
            self.__exit_critical_phase()

    def _critical_query(self, query):
        return BaseConnection.query(self, query)

    def _critical_rollback(self):
        try:
            self._direct_rollback() # pylint:disable=no-member
        finally:
            self.__exit_critical_phase()

    def _critical_commit(self):
        try:
            self._direct_commit() # pylint:disable=no-member
        finally:
            self.__exit_critical_phase()


class Connection(AbstractConnection,
                 BaseConnection):
    """
    gevent (mostly) compatible connection.

    .. rubric Blocking

    Where, when, and how long we could block the event loop is complex
    and depends on what we're doing.

    "how long" is the easy part. That's the value of the
    ``read_timeout`` constructor argument, times two (which must be >
    0; 0 means infinite). libmysql automatically includes a retry
    after an EAGAIN. If the timeout is set, then after the second try
    mysqlclient will raise ``OperationalError: 2013, Lost Connection``
    and the internal state of the connection is broken.

    This retry is done using ``vio_io_wait``, which calls
    select/epoll/kevent. We'd really like to hook into that and make
    it use the gevent loop in all cases, but I couldn't find a way to
    do that. Setting the socket to nonblocking mode does nothing
    except lead to unexpected disconnects.

    "Where" depends on the query and where it blocks. The server will
    always wait to send us the first result set, so our ``query()``
    function can successfully ``gevent_wait_read()`` to detect when
    the first batch of results arrive. If the query is going to
    produce multiple result sets (either because it contained multiple
    statements like "SELECT ...; SELECT ...;" or because it was a
    ``CALL``), the blocking (if any) will occur during
    ``next_result``. It's not possible for this object to determine
    whether the network IO done by ``next_result`` will block, or,
    indeed, if it will even do any network IO.

    You'd think it would be possible to predict whether it would need
    to block or not, or do network IO, based on whether we have
    statically determined that there's a next result set
    (``ResultObject.has_next``). That turns out not to be correct. For
    example, given a stored procedure that will return three result
    sets (the first two from queries that may return no rows), after
    executing the query, the initial ``ResultObject`` will have
    ``has_next`` of ``False`` --- but ``self.next_result()`` returns
    0, indicating that there are in fact more results (apparently
    ``mysql_more_results`` is buggy). This next ``ResultObject`` will
    have a correct value for ``has_next``, indicating a third result
    is available, but ``gevent_wait_read`` will never return,
    suggesting that the results have already been buffered.

    The upshot is that stored procedures should either return no more
    than one result set, or if there is going to be blocking, they
    should block in the first result set. We can't properly deal with
    blocking that occurs at the arbitrary time of ``next_result``
    without the possibility of deadlocking. (See
    ``GeventConnectionMixin``.)
    """

    _direct_rollback = BaseConnection.rollback
    _direct_commit = BaseConnection.commit

    def __delattr__(self, name):
        # BaseConnection has a delattr that forbids
        # all deletion. Not helpful.
        if name in self.__dict__:
            self.__dict__.pop(name)
        else:
            BaseConnection.__delattr__(self, name)

    def gevent_check_write(self):
        # This protocol is a request/reply nature, so reason dictates
        # that we will *always* be able to write to the socket: We
        # don't get control back until we've sent everything we need
        # to send.
        # Empirical testing backs that up. This is always true. So we
        # can save the syscall and just go.
        return True
